(* $Id: mimestring.mli,v 1.4 2000/05/16 22:29:12 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

(**********************************************************************)
(* Collection of auxiliary functions to parse MIME headers            *)
(**********************************************************************)


val scan_header : string -> int -> int -> ((string * string) list * int)
    (* let params, i2 = scan_header s i0 i1:
     *
     * DESCRIPTION
     *
     * Scans the MIME header that begins at position i0 in the string s
     * and that must end somewhere before position i1. It is intended
     * that in i1 the character position following the end of the body of the
     * MIME message is passed.
     * Returns the parameters of the header as (name,value) pairs (in
     * params), and in i2 the position of the character following
     * directly after the header (i.e. after the blank line separating
     * the header from the body).
     * The following normalizations have already been applied:
     * - The names are all in lowercase
     * - Newline characters (CR and LF) have been removed
     * - Whitespace at the beginning and at the end of values has been
     *   removed
     * The rules of RFC 2047 have NOT been applied.
     * The function fails if the header violates the header format
     * strongly. (Some minor deviations are tolerated, e.g. it is sufficient
     * to separate lines by only LF instead of CRLF.)
     *
     * ABOUT MIME MESSAGE FORMAT:
     *
     * This is the modern name for messages in "E-Mail format". Messages
     * consist of a header and a body; the first empty line separates both
     * parts. The header contains lines "param-name: param-value" where
     * the param-name must begin on column 0 of the line, and the ":"
     * separates the name and the value. So the format is roughly:
     *
     * param1-name: param1-value
     * ...
     * paramN-name: paramN-value
     *
     * body
     *
     * This function wants in i0 the position of the first character of
     * param1-name in the string, and in i1 the position of the character
     * following the body. It returns as i2 the position where the body
     * begins. Furthermore, in 'params' all parameters are returned that
     * exist in the header.
     *
     * DETAILS
     *
     * Note that parameter values are restricted; you cannot represent
     * arbitrary strings. The following problems can arise:
     * - Values cannot begin with whitespace characters, because there
     *   may be an arbitrary number of whitespaces between the ':' and the
     *   value.
     * - Values (and names of parameters, too) must only be formed of
     *   7 bit ASCII characters. (If this is not enough, the MIME standard
     *   knows the extension RFC 2047 that allows that header values may
     *   be composed of arbitrary characters of arbitrary character sets.)
     * - Header values may be broken into several lines, the continuation
     *   lines must begin with whitespace characters. This means that values
     *   must not contain line breaks as semantical part of the value.
     *   And it may mean that ONE whitespace character is not distinguishable
     *   from SEVERAL whitespace characters.
     * - Header lines must not be longer than 76 characters. Values that
     *   would result into longer lines must be broken into several lines.
     *   This means that you cannot represent strings that contain too few
     *   whitespace characters.
     * - Some gateways pad the lines with spaces at the end of the lines.
     *
     * This implementation of a MIME scanner tolerates a number of
     * deviations from the standard: long lines are not rejected; 8 bit
     * values are accepted; lines may be ended only with LF instead of
     * CRLF.
     * Furthermore, header values are transformed:
     * - leading and trailing spaces are always removed
     * - CRs and LFs are deleted; it is guaranteed that there is at least
     *   one space or tab where CR/LFs are deleted.
     * Last but not least, the names of the header values are converted
     * to lowercase; MIME specifies that they are case-independent.
     *
     * COMPATIBILITY WITH THE STANDARD
     *
     * This function can parse all MIME headers that conform to RFC 822.
     * But there may be still problems, as RFC 822 allows some crazy
     * representations that are actually not used in practice.
     * In particular, RFC 822 allows it to use backslashes to "indicate"
     * that a CRLF sequence is semantically meant as line break. As this
     * functions delete CRLFs, it is not possible to recognize such
     * indicators in the result of the function.
     *)
    (* TODO: functions to support RFC 2047 *)


type s_token =
    Atom of string
  | QString of string
  | Control of char
  | Special of char


type s_option =
    No_backslash_escaping
      (* Do not handle backslashes in quoted string and comments as escape
       * characters; backslashes are handled as normal characters.
       * For example: "C:\dir\file" will be returned as
       * QString "C:\dir\file", and not as QString "C:dirfile".
       * - This is a common error in many MIME implementations.
       *)


val scan_structured_value : string -> char list -> s_option list -> s_token list
    (* let seq = scan_structured_value s options specials:
     *
     * Scans the structured value, and returns a sequence of words,
     * special characters, and control characters.
     *
     * WHAT ARE STRUCTURED VALUES?
     *
     * RFC 822 (together with some other RFCs) defines lexical rules
     * how formal MIME header values should be divided up into tokens. Formal
     * MIME headers are those headers that are formed according to some
     * grammar, e.g. mail addresses or MIME types.
     *    Some of the characters separate phrases of the value; these are
     * the "special" characters. For example, '@' is normally a special
     * character for mail addresses, because it separates the user name
     * from the domain name. RFC 822 defines a fixed set of special
     * characters, but other RFCs use different sets. Because of this,
     * this function allows it to configure the set of special characters.
     *    Every sequence of characters may be embraced by double quotes,
     * which means that the sequence is meant as literal data item;
     * special characters are not recognized inside a quoted string. You may
     * use the backslash to insert any character (including double quotes)
     * verbatim into the quoted string (e.g. "He said: \"Give it to me!\"").
     *    Structured values may contain comments. The beginning of a comment
     * is indicated by '(', and the end by ')'. Comments may be nested.
     * Again, the backslash allows verbatim insertion of characters. A
     * comment counts as if a space character were written instead of it.
     *    Control characters are the ASCII characters 0 to 31, and 127.
     * RFC 822 demands that MIME headers are 7 bit ASCII strings. Because
     * of this, this function also counts the characters 128 to 255 as
     * control characters.
     *    Every character sequence not falling in one of the above categories
     * is an atom (a sequence of non-special and non-control characters).
     *
     * INTERFACE:
     *
     * The character list 'specials' specifies the set of special characters.
     * If ' ' is member of this list, spaces and comments are returned as
     * Special ' '; otherwise spaces and comments are ignored.
     * If '\t' is member of this list, tabs are returned as Special '\t';
     * otherwise tabs are ignored.
     * If '"' is member of this list, quoted strings are not recognized, and
     * double quotes are returned as Special '"'.
     * If '(' is member of this list, comments are not recognized, and
     * '(' characters are returned as Special '('.
     * All other characters c in the list 'specials' are simply returned as
     * Special c.
     *
     * Atoms are returned as Atom s where s is a longest sequence of
     * atomic characters.
     * If recognized, quoted strings are returned as QString s, where
     * s is the string without the embracing quotes, and with already
     * resolved backslash quotations.
     * Control characters c are returned as Control c.
     * If recognized, comments are not returned as such; but if ' ' is a member
     * of 'specials', at least a Special ' ' is returned instead of the comment.
     *
     * EXAMPLES
     *
     * scan_structured_value "user@domain.com" [ '@'; '.' ] []
     *   = [ Atom "user"; Special '@'; Atom "domain"; Special '.'; Atom "com" ]
     *
     * scan_structured_value "user @ domain . com" [ '@'; '.' ] []
     *   = [ Atom "user"; Special '@'; Atom "domain"; Special '.'; Atom "com" ]
     *
     * scan_structured_value "user(Do you know him?)@domain.com" [ '@'; '.' ] []
     *   = [ Atom "user"; Special '@'; Atom "domain"; Special '.'; Atom "com" ]
     *
     * scan_structured_value "user @ domain . com" [ '@'; '.'; ' ' ] []
     *   = [ Atom "user"; Special ' '; Special '@'; Special ' '; Atom "domain";
     *       Special ' '; Special '.'; Special ' '; Atom "com" ]
     *
     * scan_structured_value "user(Do you know him?)@domain.com" ['@'; '.'; '(']
     *     []
     *   = [ Atom "user"; Special '('; Atom "Do"; Atom "you"; Atom "know";
     *       Atom "him?)"; Special '@'; Atom "domain"; Special '.'; Atom "com" ]
     *
     * scan_structured_value "\"My.name\"@domain.com" [ '@'; '.' ] []
     *   = [ QString "My.name"; Special '@'; Atom "domain"; Special '.';
     *       Atom "com" ]
     *)

val specials_rfc822 : char list
val specials_rfc2045 : char list
    (* The sets of special characters defined by the RFCs 822 and 2045. *)

val scan_value_with_parameters : string -> s_option list ->
                                   (string * (string * string) list)
    (* let name, params = scan_value_with_parameters s options:
     * Scans phrases like
     *    name ; p1=v1 ; p2=v2 ; ...
     * The scan is done with the set of special characters [';', '='].
     *)

val scan_mime_type : string -> s_option list ->
                       (string * (string * string) list)
    (* let name, params = scan_mime_type s options:
     * Scans MIME types like
     *    text/plain; charset=iso-8859-1
     * The name of the type and the names of the parameters are converted
     * to lower case.
     *)

val scan_multipart_body : string -> int -> int -> string ->
                            ((string * string) list * string) list
    (* let [params1, value1; params2, value2; ...]
     *   = scan_multipart_body s i0 i1 b
     *
     * Scans the string s that is the body of a multipart message.
     * The multipart message begins at position i0 in s and i1 the position
     * of the character following the message. In b the boundary string
     * must be passed (this is the "boundary" parameter of the multipart
     * MIME type, e.g. multipart/mixed;boundary="some string" ).
     *     The return value is the list of the parts, where each part
     * is returned as pair (params, value). The left component params
     * is the list of name/value pairs of the header of the part. The
     * right component is the RAW content of the part, i.e. if the part
     * is encoded ("content-transfer-encoding"), the content is returned
     * in the encoded representation. The caller must himself decode
     * the content.
     *     The material before the first boundary and after the last
     * boundary is not returned.
     *
     * MULTIPART MESSAGES
     *
     * The MIME standard defines a way to group several message parts to
     * a larger message (for E-Mails this technique is known as "attaching"
     * files to messages); these are the so-called multipart messages.
     * Such messages are recognized by the major type string "multipart",
     * e.g. multipart/mixed or multipart/form-data. Multipart types MUST
     * have a boundary parameter because boundaries are essential for the
     * representation.
     *    Multipart messages have a format like
     *
     * ...Header...
     * Content-type: multipart/xyz; boundary="abc"
     * ...Header...
     *
     * Body begins here ("prologue")
     * --abc
     * ...Header part 1...
     *
     * ...Body part 1...
     * --abc
     * ...Header part 2...
     *
     *
     * ...Body part 2
     * --abc
     * ...
     * --abc--
     * Epilogue
     *
     * The parts are separated by boundary lines which begin with "--" and
     * the string passed as boundary parameter. (Note that there may follow
     * arbitrary text on boundary lines after "--abc".) The boundary is
     * chosen such that it does not occur as prefix of any line of the
     * inner parts of the message.
     *     The parts are again MIME messages, with header and body. Note
     * that it is explicitely allowed that the parts are even multipart
     * messages.
     *     The texts before the first boundary and after the last boundary
     * are ignored.
     *     Note that multipart messages as a whole MUST NOT be encoded.
     * Only the PARTS of the messages may be encoded (if they are not
     * multipart messages themselves).
     *
     * Please read RFC 2046 if want to know the gory details of this
     * brain-dead format.
     *)

val scan_multipart_body_and_decode : string -> int -> int -> string ->
                                        ((string * string) list * string) list
    (* Same as scan_multipart_body, but decodes the bodies of the parts
     * if they are encoded using the methods "base64" or "quoted printable".
     * Fails, if an unknown encoding is used.
     *)

val scan_multipart_body_from_netstream
    : Netstream.t ->
      string ->
      ((string * string) list -> 'a) ->
      ('a -> Netstream.t -> int -> int -> unit) ->
      ('a -> unit) ->
      unit
    (* scan_multipart_body_from_netstream s b create add stop:
     *
     * Reads the MIME message from the netstream s block by block. The
     * parts are delimited by the boundary b.
     *
     * Once a new part is detected and begins, the function 'create' is
     * called with the MIME header as argument. The result p of this function
     * may be of any type.
     *
     * For every chunk of the part that is being read, the function 'add'
     * is invoked: add p s k n.
     * Here, p is the value returned by the 'create' invocation for the
     * current part. s is the netstream. The current window of s contains
     * the read chunk completely; the chunk begins at position k of the
     * window (relative to the beginning of the window) and has a length
     * of n bytes.
     *
     * When the part has been fully read, the function 'stop' is
     * called with p as argument.
     *
     * That means, for every part the following is executed:
     * - let p = create h
     * - add p s k1 n1
     * - add p s k2 n2
     * - ...
     * - add p s kN nN
     * - stop p
     *
     * IMPORTANT PRECONDITION:
     * - The block size of the netstream s must be at least
     *   String.length b + 3
     *
     * EXCEPTIONS:
     * - Exceptions can happen because of ill-formed input, and within
     *   the callbacks of the functions 'create', 'add', 'stop'.
     * - If the exception happens while part p is being read, and the
     *   'create' function has already been called (successfully), the
     *   'stop' function is also called (you have the chance to close files).
     *)


(* THREAD-SAFETY:
 * Most functions of this module are NOT thread-safe.
 *)

(* ======================================================================
 * History:
 *
 * $Log: mimestring.mli,v $
 * Revision 1.4  2000/05/16 22:29:12  gerd
 * 	New "option" arguments specifying the level of MIME
 * compatibility.
 *
 * Revision 1.3  2000/04/15 13:09:01  gerd
 * 	Implemented uploads to temporary files.
 *
 * Revision 1.2  2000/03/02 01:15:30  gerd
 * 	Updated.
 *
 * Revision 1.1  2000/02/25 15:21:12  gerd
 * 	Initial revision.
 *
 *
 *)
